/*
 *  for new GOGO-no-coda (1999/10/7)
 *  modified by shigeo
 */

/*
 *	Version numbering for LAME.
 *
 *	Copyright (c) 1999 A.L. Faber
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <memory.h>
#include <float.h>
#include <math.h>
#include "vbrtag.h"
#include "musenc.h"
#include "bitstrem.h"

static int SizeOfEmptyFrame[2][2]=
{
	{32,17},
	{17,9},
};

static BYTE pbtStreamBuffer[216];   
static long g_Position[NUMTOCENTRIES];
static int nZeroStreamSize=0;
static int TotalFrameSize=0;
static char	VBRTag[]={"Xing"};

static int* pVbrFrames=NULL;
static int nVbrNumFrames=0;
static int nVbrFrameBufferSize=0;

/****************************************************************************
 * AddVbrFrame: Add VBR entry, used to fill the VBR the TOC entries
 * Paramters:
 *	nStreamPos: how many bytes did we write to the bitstream so far
 *				(in Bytes NOT Bits)
 ****************************************************************************
*/
void AddVbrFrame(int nStreamPos)
{
	// Simple exponential growing buffer
	if (pVbrFrames==NULL || nVbrFrameBufferSize==0)
	{
		// Start with 100 frames
		nVbrFrameBufferSize=100;

		// Allocate them 
		pVbrFrames=(int*)malloc(nVbrFrameBufferSize*sizeof(int));
	}

	// Is buffer big enough to store this new frame
	if (nVbrNumFrames==nVbrFrameBufferSize)
	{
		// Guess not, double th e buffer size
		nVbrFrameBufferSize*=2;

		// Allocate new buffer
		pVbrFrames=(int*)realloc(pVbrFrames,nVbrFrameBufferSize*sizeof(int));
	}

	// Store values
	pVbrFrames[nVbrNumFrames++]=nStreamPos;
}

/*-------------------------------------------------------------*/
static int ExtractI4(unsigned char *buf)
{
	int x;
	// big endian extract
	x = buf[0];
	x <<= 8;
	x |= buf[1];
	x <<= 8;
	x |= buf[2];
	x <<= 8;
	x |= buf[3];
	return x;
}

static void CreateI4(unsigned char *buf, int nValue)
{
	// big endian create
	buf[0]=(nValue>>24)&0xff;
	buf[1]=(nValue>>16)&0xff;
	buf[2]=(nValue>> 8)&0xff;
	buf[3]=(nValue    )&0xff;
}

/****************************************************************************
 * InitVbrTag: Initializes the header, and write empty frame to stream
 * Paramters:
 *				fpStream: pointer to output file stream
 *				nVersion: 0= MPEG1 1=MPEG2
 *				nMode	: Channel Mode: 0=STEREO 1=JS 2=DS 3=MONO
 ****************************************************************************
*/
int InitVbrTag(int nVersion, int nMode, int SampIndex)
{
	int i;

	// Clear Frame position array variables
	pVbrFrames=NULL;
	nVbrNumFrames=0;
	nVbrFrameBufferSize=0;

	// Clear struct
	memset(g_Position,0x00,sizeof(g_Position));

	// Clear stream buffer
	memset(pbtStreamBuffer,0x00,sizeof(pbtStreamBuffer));

	// Set TOC values to 255
	for (i=0;i<NUMTOCENTRIES;i++)
	{
		g_Position[i]=-1;
	}



	// Reserve the proper amount of bytes
	if (nMode==3)
	{
		nZeroStreamSize=SizeOfEmptyFrame[nVersion][1]+4;
	}
	else
	{
		nZeroStreamSize=SizeOfEmptyFrame[nVersion][0]+4;
	}


	// Xing VBR pretends to be a 48kbs layer III frame.  (at 44.1kHz).
        // what do they do at 48kHz?  frame not big enough for Xing header.
	// let's always embed Xing header inside a 64kbs layer III frame.  
	// this gives us enough room for a LAME version string too.
	// size determined by sampling frequency (MPEG1)
	// 32kHz:    216 bytes@48kbs    288bytes@ 64kbs    
	// 44.1kHz:  156 bytes          208bytes@64kbs     (+1 if padding = 1)
	// 48kHz:    144 bytes          192
	// 
	// MPEG 2 values are half the above.  Need to go to a 128kbs frame
	// to have room for the table of contents.  not yet supported.
	if (nVersion == 0 ) {
	  int tot;
#define XING64
#ifdef XING64
	  int framesize[3]={208,192,288};
#else
	  int framesize[3]={156,144,216}; 
#endif
	  if (SampIndex==3) {
	    fprintf(stderr,"illegal sampling frequency index\n");
           TERM( ME_INTERNALERROR );//exit(-1);
	  }
	  TotalFrameSize= framesize[SampIndex];
	  tot = (nZeroStreamSize+VBRHEADERSIZE);
#ifdef XING64
	  tot += 20;  /* extra 20 bytes for LAME & version string */
#endif
	  if (TotalFrameSize < tot ) {
	    fprintf(stderr,"Xing VBR header problem...use -t\n");
           TERM( ME_INTERNALERROR );//exit(-1);
	  }
	}
	if (nVersion == 1) {
	  fprintf(stderr,"MPEG 2 Xing VBR header not supported. use -t\n");
           TERM( ME_INTERNALERROR );//exit(-1);
	}


	// Put empty bytes into the bitstream
	for (i=0;i<TotalFrameSize;i++)
	{
		// Write a byte to the bitstream
             putbits(0,8);
	}

	// Success
	return 0;
}



/****************************************************************************
 * PutVbrTag: Write final VBR tag to the file
 * Paramters:
 *				lpszFileName: filename of MP3 bit stream
 *				nVbrScale	: encoder quality indicator (0..100)
 ****************************************************************************
*/
int PutVbrTag(char* lpszFileName,int nVbrScale)
{
	int			i;
	long		lFileSize=0;
	int			nStreamIndex=0;
	char abyte;
	BYTE		btToc[NUMTOCENTRIES];
	FILE*		fpStream=NULL;
	char str1[80];


	if (nVbrNumFrames==0 || pVbrFrames==NULL)
		return -1;

	// Open the bitstream again
	fpStream=fopen(lpszFileName,"rb+");

	// Assert stream is valid
	if (fpStream==NULL)
		return -1;

	// Clear stream buffer
	memset(pbtStreamBuffer,0x00,sizeof(pbtStreamBuffer));

	// Seek to end of file
	fseek(fpStream,0,SEEK_END);

	// Get file size
	lFileSize=ftell(fpStream);
	
		// Seek to first real frame
	fseek(fpStream,TotalFrameSize,SEEK_SET);

	// Read the header (first valid frame)
	fread(pbtStreamBuffer,4,1,fpStream);

	/* the default VBR header.  48kbs layer III, no padding, no crc */
	/* but sampling freq, mode andy copyright/copy protection taken */
	/* from first valid frame */
	pbtStreamBuffer[0]=(char) 0xff;    
	pbtStreamBuffer[1]=(char) 0xfb;    
	abyte = pbtStreamBuffer[2] & (char) 0x0c;   
#ifdef XING64
	pbtStreamBuffer[2]=(char) 0x50 | abyte;     /* 64kbs frame */
#else
	pbtStreamBuffer[2]=(char) 0x30 | abyte;  /* 48kbs frame */
#endif

	// Seek to the beginning of the stream
	fseek(fpStream,0,SEEK_SET);

	// Clear all TOC entries
	memset(btToc,0,sizeof(btToc));

	// Set unassigned toc value with its neighbor value
	for (i=0;i<nVbrNumFrames;i++)
	{
		// Calculate percentage
		int nPercent= i*(NUMTOCENTRIES-1)/nVbrNumFrames;

		// Calculate relative file postion, normalized to 0..255
		float fRelStreamPos=255*(float)pVbrFrames[i]/(float)lFileSize;

		// Just to be safe
		if (nPercent>=NUMTOCENTRIES ) nPercent=NUMTOCENTRIES-1;
		if (fRelStreamPos>255) fRelStreamPos=255;

		// Assign toc entry value
		btToc[nPercent]=(BYTE)fRelStreamPos;
	}

	// Fix up empty fields (if we have less than 100 frames)
	for (i=1;i<NUMTOCENTRIES ;i++)
	{
		if (btToc[i]==0)
			btToc[i]=btToc[i-1];
	}


	// Start writing the tag after the zero frame
	nStreamIndex=nZeroStreamSize;

	// Put Vbr tag
	pbtStreamBuffer[nStreamIndex++]=VBRTag[0];
	pbtStreamBuffer[nStreamIndex++]=VBRTag[1];
	pbtStreamBuffer[nStreamIndex++]=VBRTag[2];
	pbtStreamBuffer[nStreamIndex++]=VBRTag[3];

	// Put header flags
	CreateI4(&pbtStreamBuffer[nStreamIndex],FRAMES_FLAG+BYTES_FLAG+TOC_FLAG+VBR_SCALE_FLAG);
	nStreamIndex+=4;

	// Put Total Number of frames
	CreateI4(&pbtStreamBuffer[nStreamIndex],nVbrNumFrames);
	nStreamIndex+=4;

	// Put Total file size
	CreateI4(&pbtStreamBuffer[nStreamIndex],lFileSize);
	nStreamIndex+=4;

	// Put TOC
	memcpy(&pbtStreamBuffer[nStreamIndex],btToc,sizeof(btToc));
	nStreamIndex+=sizeof(btToc);

	// Put VBR SCALE
	CreateI4(&pbtStreamBuffer[nStreamIndex],nVbrScale);
	nStreamIndex+=4;

	// Put LAME id
        sprintf(str1,"MPGE");
	strncpy((char *)&pbtStreamBuffer[nStreamIndex],str1,(size_t) 20);
	nStreamIndex+=20;


#ifdef DEBUG_VBRTAG
{
	VBRTAGDATA TestHeader;
	GetVbrTag(&TestHeader,pbtStreamBuffer);
}
#endif

	// Put it all to disk again
	if (fwrite(pbtStreamBuffer,TotalFrameSize,1,fpStream)!=1)
	{
		// Could not write header
		return -1;
	}

	fclose(fpStream);

	free(pVbrFrames);
	pVbrFrames=NULL;

	return 0;       // success
}

// kazumi-t@ss.iij4u.or.jp
int PutVbrTagStream(char* pBuffer,int nVbrScale, long lFileSize, int SampIndex)
{
	int			i;
	int			nStreamIndex=0;
	char abyte;
	BYTE		btToc[NUMTOCENTRIES];
	char str1[80];


	if (nVbrNumFrames==0 || pVbrFrames==NULL)
		return -1;

	// Clear stream buffer
	memset(pbtStreamBuffer,0x00,sizeof(pbtStreamBuffer));

	/* the default VBR header.  48kbs layer III, no padding, no crc */
	/* but sampling freq, mode andy copyright/copy protection taken */
	/* from first valid frame */
	pbtStreamBuffer[0]=(char) 0xff;    
	pbtStreamBuffer[1]=(char) 0xfb;    
	abyte = pbtStreamBuffer[2] & (char) 0x0c;   
#ifdef XING64
	pbtStreamBuffer[2]=(char) 0x50 | abyte | (SampIndex << 2);     /* 64kbs frame */
#else
	pbtStreamBuffer[2]=(char) 0x30 | abyte | (SampIndex << 2);  /* 48kbs frame */
#endif

	// Clear all TOC entries
	memset(btToc,0,sizeof(btToc));

	// Set unassigned toc value with its neighbor value
	for (i=0;i<nVbrNumFrames;i++)
	{
		// Calculate percentage
		int nPercent= i*(NUMTOCENTRIES-1)/nVbrNumFrames;

		// Calculate relative file postion, normalized to 0..255
		float fRelStreamPos=255*(float)pVbrFrames[i]/(float)lFileSize;

		// Just to be safe
		if (nPercent>=NUMTOCENTRIES ) nPercent=NUMTOCENTRIES-1;
		if (fRelStreamPos>255) fRelStreamPos=255;

		// Assign toc entry value
		btToc[nPercent]=(BYTE)fRelStreamPos;
	}

	// Fix up empty fields (if we have less than 100 frames)
	for (i=1;i<NUMTOCENTRIES ;i++)
	{
		if (btToc[i]==0)
			btToc[i]=btToc[i-1];
	}


	// Start writing the tag after the zero frame
	nStreamIndex=nZeroStreamSize;

	// Put Vbr tag
	pbtStreamBuffer[nStreamIndex++]=VBRTag[0];
	pbtStreamBuffer[nStreamIndex++]=VBRTag[1];
	pbtStreamBuffer[nStreamIndex++]=VBRTag[2];
	pbtStreamBuffer[nStreamIndex++]=VBRTag[3];

	// Put header flags
	CreateI4(&pbtStreamBuffer[nStreamIndex],FRAMES_FLAG+BYTES_FLAG+TOC_FLAG+VBR_SCALE_FLAG);
	nStreamIndex+=4;

	// Put Total Number of frames
	CreateI4(&pbtStreamBuffer[nStreamIndex],nVbrNumFrames);
	nStreamIndex+=4;

	// Put Total file size
	CreateI4(&pbtStreamBuffer[nStreamIndex],lFileSize);
	nStreamIndex+=4;

	// Put TOC
	memcpy(&pbtStreamBuffer[nStreamIndex],btToc,sizeof(btToc));
	nStreamIndex+=sizeof(btToc);

	// Put VBR SCALE
	CreateI4(&pbtStreamBuffer[nStreamIndex],nVbrScale);
	nStreamIndex+=4;

	// Put LAME id
        sprintf(str1,"MPGE");
	strncpy((char *)&pbtStreamBuffer[nStreamIndex],str1,(size_t) 20);
	nStreamIndex+=20;

	memcpy(pBuffer, pbtStreamBuffer, TotalFrameSize);

	// Save to delete the frame buffer
	free(pVbrFrames);
	pVbrFrames=NULL;

	return TotalFrameSize;
}
